"""Automatic source link generation for mkdocstrings."""
import shutil
from abc import ABC, abstractmethod, abstractproperty
from pathlib import Path
from typing import Dict, List, Union

from mkdocstrings_sourcelink.toolbox import Utilities


class BuilderMkDoc(ABC):
    """BuilderMkDoc is the Metaclass of `MkDocGenerator`.

    Args:
        ABC (class): Helper class that provides a standard way to create an ABC using
             inheritance.
    """

    @abstractmethod
    def render_to_markdown(self, element: str) -> str:
        """Abstract the internal method of `render_to_markdown`."""

    @abstractmethod
    def initialize_generate(self) -> None:
        """Abstract the internal method of `initialize_generate`."""

    @abstractmethod
    def generate_docs(self) -> None:
        """Abstract the internal method of `generate_docs`."""

    @abstractmethod
    def generate_static(self) -> None:
        """Abstract the internal method of `generate_static`."""

    @abstractproperty
    def generate(self) -> None:
        """Abstract the property of `generate`."""


class MkDocGenerator(Utilities, BuilderMkDoc):
    """The `MkDocGenerator` generates the documentation with the links to the source code.

    Args:
        Utilities (class): [description]
        BuilderMkDoc (class): Builder class of the abstract methods and property of
             `MkDocGenerator`.
    """

    def __init__(
        self,
        dest_dir: Union[str, Path],
        documentation: Dict[str, Dict[str, List[str]]],
        project_url: Union[str, Dict[str, str]] = None,
        template_dir: Union[str, Path] = None,
        example_dir: Union[str, Path] = None,
        markdown_files: Dict[str, Dict[str, List[str]]] = None,
        titles_size: str = "#",
        underline_title: bool = False,
        source: str = "**source code**",
    ) -> None:
        """Generates the documentation via `MkDocGenerator` with the links to the source.

        Args:
            dest_dir (Union[str, Path]): Destination of the generated documentation.
            documentation (Dict[str, Dict[str, List[str]]], optional): A nested dictionary with the
                 page title, the page filename, and the functions /classes / methods names of the
                 page. The dictionary should look like:
                 `pages = {'page_title':{'filename.md': ['package.module.function']}}`.
            project_url (Union[str, Dict[str, str]], optional): The URL, where the project is
                 hosted and where it should be linked to (including branch and storage).
            template_dir (Union[str, Path], optional): Directory of template files. If template has
                 to be automatically filled out, then the keyword **{{autogenerated}}** has to be
                 used.
            example_dir (Union[str, Path], optional): Directory of example files, especially
                 suitable for `Jupyter-Notebooks`.
            markdown_files (Dict[str, Dict[str, List[str]]], optional): A nested dictionary with
                 the page title, the page filename, and the link to already existing markdown files
                 like **README.md** or **LICENSE**. The dictionary should look like:
                 `pages = {'page_title':{'filename.md': ['existing_file.md']}}`.
            titles_size (str, optional): Defines the initial title size for the headings, which are
                 scaled down by the factor of one "#".
            underline_title (bool, optional): Underline titles of classes, functions, and methods.
            source (str, optional): Name of the source link.

        !!! info "About *documentation*"
            In terms of linking to the project_url, it has to be separated between *specific* and
             *general* linking. If functions or method has to be separately linked, then each
             module of the class `Documentation` has to be called individually like:
            ```python
                pages = {
                    "Documentation": {
                        "auto_generator.md": [
                            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.__init__",
                            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.render_to_markdown",
                            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.initialize_generate",
                            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.generate_docs",
                            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.generate_static",
                            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.generate",
                        ]
                    }
                }
            ```
            If functions or method has to be generally linked, then just call the class
             `Documentation` is enough. It will be only generated one single link to the start of
              the class.
              ```python
                  pages = {
                      "Documentation": {
                          "auto_generator.md": [
                              "mkdocstrings_sourcelink.auto_generator.MkDocGenerator",
                      }
                  }
              ```
            For more information, please check: **mkdocstrings**
            https://github.com/pawamoy/mkdocstrings

        !!! tip "About *source*"
            Instead of using a string for `source = "**source code**"`, icons can be used instead
             or as combination of string + icon(s) like.
            ```python
                source = ":material-github::material-source-branch: source-code"
            ```
            In case of using material-icons, please check https://pictogrammers.github.io/@mdi/font/5.4.55/
            and replace `mdl` by `material`.

        !!! warning "About *project_url*"
            Keep in mind, that the name of the branch like *master*, *main*, or *dev*, as well as,
             the name of the storage like *blob* for GitHub has to be included.
        """
        self.dest_dir = Path(dest_dir)
        self.documentation = documentation
        self.project_url = project_url
        self.template_dir = Utilities.return_as_Path(template_dir)
        self.example_dir = Utilities.return_as_Path(example_dir)
        self.markdown_files = markdown_files
        self.titles_size = titles_size
        self.underline_title = underline_title
        self.source = source

    def render_to_markdown(self, element: str) -> str:
        """Rendering the element path to mkdocstrings.

        Args:
            element (str): String of they python class, function, or method, which has to be
                 converted to a string in the mkdocstrings format.

        Returns:
            str: Return of the initial string which looks like
                 `mkdocstrings_sourcelink.auto_generator.MkDocGenerator.__init__` into a markdown
                 conformed string.
        """
        object_ = Utilities.import_object(element)
        subblocks = []
        if self.project_url:
            subblocks.append(
                Utilities.make_source_link(object_, self.project_url, self.source)
            )

        subblocks.append(
            Utilities.make_title(object_, self.titles_size, self.underline_title)
        )
        subblocks.append(Utilities.element_to_mkdocstrings(element, self.titles_size))
        return "\n\n".join(subblocks) + "\n\n"

    def initialize_generate(self) -> None:
        """Initialization of the auto documentation generatorion.

        1. Firs removing a possible existing target directory (`dest_dir`).
        2. Copy templates from the template directory the target directory(`dest_dir`).
        3. Copy example from the example directory the target directory (`dest_dir`).
        """
        if self.dest_dir.exists():
            print(f"Cleaning up existing sources directory '{self.dest_dir}'.")
            shutil.rmtree(self.dest_dir)

        if self.template_dir:
            print(
                f"...copying existing sources directory '{self.template_dir}' to '{self.dest_dir}'."
            )
            shutil.copytree(self.template_dir, self.dest_dir)

        if self.example_dir:
            print(
                f"...copying existing sources directory '{self.example_dir}' to '{self.dest_dir}'."
            )
            shutil.copytree(self.example_dir, self.dest_dir)

    def generate_docs(self) -> None:
        """Generated *dynamic* documentation based on calling the elements via dictionary."""
        for title, documentation in self.documentation.items():
            markdown_text = f"{self.titles_size} {title}\n\n---\n\n"
            for file_path, elements in documentation.items():
                markdown_text += "".join(
                    self.render_to_markdown(element) for element in elements
                )
            Utilities.insert_in_file(markdown_text, self.dest_dir.joinpath(file_path))

    def generate_static(self) -> None:
        """Generate *static* documentation based on existing markdown files."""
        if self.markdown_files:
            for _, markdown_files in self.markdown_files.items():
                markdown_text = ""
                for file_path, elements in markdown_files.items():
                    markdown_text += "".join(
                        Path(element).read_text() for element in elements
                    )
                Utilities.insert_in_file(
                    markdown_text, self.dest_dir.joinpath(file_path)
                )

    @property
    def generate(self) -> None:
        """The property `generate` of `MkDocGenerator` creates the final markdown files.

        !!! note "Example: How to use"

            ```python
            >>> from pathlib import Path
            # Import mkdocstrings-sourcelink
            >>> from mkdocstrings_sourcelink import MkDocGenerator
            # Define the dictionary for the documentation
            >>> pages = {
                "Documentation": {
                    "auto_generator.md": [
                        "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.__init__",
                        "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.render_to_markdown",
                        "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.initialize_generate",
                        "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.generate_docs",
                        "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.generate_static",
                        "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.generate",
                    ]
                },
                "Tools": {
                    "toolbox.md": [
                        "mkdocstrings_sourcelink.toolbox.Utilities.insert_in_file",
                        "mkdocstrings_sourcelink.toolbox.Utilities.element_to_mkdocstrings",
                        "mkdocstrings_sourcelink.toolbox.Utilities.make_source_link",
                        "mkdocstrings_sourcelink.toolbox.Utilities.make_title",
                        "mkdocstrings_sourcelink.toolbox.Utilities.import_object",
                        "mkdocstrings_sourcelink.toolbox.Utilities.return_as_Path",
                    ],
                },
            }
            # Define the dictionary for the importing existing markdown files.
            >>> markdown_files = {
                "HOME": {"index.md": ["../README.md"]},
                "CONTRIBUTING": {"contributing.md": ["../CONTRIBUTING.md"]},
                "LICENSE": {"license.md": ["../LICENSE"]},
            }
            >>> root = Path(__file__).resolve().parents[1]
            # Make use of the generate poperty of MkDocGenerator
            >>> MkDocGenerator(
                root / "docs" / "src",
                pages,
                "https://github.com/AI2Business/mkdocstrings-sourcelink/blob/main",
                markdown_files=markdown_files,
                underline_title=True,
                source=":material-github::material-source-branch:",
            ).generate
            >>> ...
            ```
        """
        self.initialize_generate()
        self.generate_docs()
        self.generate_static()



from pathlib import Path

import pytest

from mkdocstrings_sourcelink import MkDocGenerator

pages = {
    "Documentation": {
        "auto_generator.md": [
            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.__init__",
            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.render_to_markdown",
            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.initialize_generate",
            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.generate_docs",
            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.generate_static",
            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator.generate",
        ]
    },
    "Tools": {
        "toolbox.md": [
            "mkdocstrings_sourcelink.toolbox.Utilities.insert_in_file",
            "mkdocstrings_sourcelink.toolbox.Utilities.element_to_mkdocstrings",
            "mkdocstrings_sourcelink.toolbox.Utilities.make_source_link",
            "mkdocstrings_sourcelink.toolbox.Utilities.make_title",
            "mkdocstrings_sourcelink.toolbox.Utilities.import_object",
            "mkdocstrings_sourcelink.toolbox.Utilities.return_as_Path",
        ],
    },
    "API": {
        "api.md": [
            "mkdocstrings_sourcelink.auto_generator.MkDocGenerator",
            "mkdocstrings_sourcelink.toolbox.Utilities",
        ],
    },
}
markdown_files = {"HOME": {"index.md": ["CODE_OF_CONDUCT.md"]}}
root = Path(__file__).resolve().parents[1]
_dir = root / "test" / "docs" / "src"
_tmp_dir = root / "test" / "docs" / "tmp"
_tmp_dir.mkdir(parents=True, exist_ok=True)


def test_basic_1():

    MkDocGenerator(
        _dir,
        pages,
        "https://github.com/AI2Business/mkdocstrings-sourcelink",
    ).generate
    assert len(list(_dir.glob("*.md"))) == 3


def test_basic_2():

    MkDocGenerator(
        _dir,
        pages,
        "https://github.com/AI2Business/mkdocstrings-sourcelink",
        underline_title=True,
    ).generate
    assert len(list(_dir.glob("*.md"))) == 3


def test_dict_url():

    MkDocGenerator(
        _dir,
        pages,
        {
            "mkdocstrings_sourcelink": "https://github.com/AI2Business/mkdocstrings-sourcelink"
        },
    ).generate
    assert len(list(_dir.glob("*.md"))) == 3


def test_markdown():

    MkDocGenerator(
        _dir,
        pages,
        "https://github.com/AI2Business/mkdocstrings-sourcelink",
        markdown_files=markdown_files,
    ).generate
    assert len(list(_dir.glob("*.md"))) == 4


def test_tmp():

    MkDocGenerator(
        _dir,
        pages,
        "https://github.com/AI2Business/mkdocstrings-sourcelink",
        template_dir=_tmp_dir,
    ).generate
    assert len(list(_dir.glob("*.md"))) == 3


def test_example():

    MkDocGenerator(
        _dir,
        pages,
        "https://github.com/AI2Business/mkdocstrings-sourcelink",
        example_dir=_tmp_dir,
    ).generate
    assert len(list(_dir.glob("*.md"))) == 3


def test_tmp_failed():
    with open(_tmp_dir / "template_test.md", "w") as f:
        f.write("hallo")
    with pytest.raises(Exception):
        MkDocGenerator(
            _dir,
            {
                "TemplateTest": {
                    "template_test.md": [
                        "mkdocstrings_sourcelink.auto_generator.MkDocGenerator",
                    ]
                }
            },
            "https://github.com/AI2Business/mkdocstrings-sourcelink",
            template_dir=_tmp_dir,
        ).generate


def test_tmp_run():
    with open(_tmp_dir / "template_test.md", "w") as f:
        f.write("{{autogenerated}}")
    MkDocGenerator(
        _dir,
        {
            "TemplateTest": {
                "template_test.md": [
                    "mkdocstrings_sourcelink.auto_generator.MkDocGenerator",
                ]
            }
        },
        "https://github.com/AI2Business/mkdocstrings-sourcelink",
        template_dir=_tmp_dir,
    ).generate
    assert len(list(_dir.glob("*.md"))) == 1



"""Tools for the automatic source link generation for mkdocstrings."""
import importlib
import inspect
from abc import ABC, abstractstaticmethod
from pathlib import Path
from typing import Dict, Optional, Union


class BuilderUtilities(ABC):
    """BuilderUtilities is the Metaclass of Utilities.

    The Utilities are essential for building the mkdocstrings and hyperlinks to the source code.

    Args:
        ABC (class): Helper class that provides a standard way to create an ABC using
             inheritance.
    """

    @abstractstaticmethod
    def insert_in_file(markdown_text: str, file_path: Path) -> None:
        """Abstract staticmethod of `insert_in_file`."""

    @abstractstaticmethod
    def element_to_mkdocstrings(element: str, titles_size: str) -> str:
        """Abstract staticmethod of `element_to_mkdocstrings`."""

    @abstractstaticmethod
    def make_source_link(
        cls: classmethod,
        project_url: Union[str, Dict[str, str]],
        source: str = "**source code**",
    ) -> str:
        """Abstract staticmethod of `make_title`."""

    @abstractstaticmethod
    def make_title(cls: classmethod, titles_size: str, underline_title: bool) -> str:
        """Abstract staticmethod of `import_object`."""

    @abstractstaticmethod
    def return_as_Path(path: str = None) -> Optional[Path]:
        """Abstract staticmethod of `return_as_Path`."""


class Utilities(BuilderUtilities):
    """The `Utilities` build the mkdocstrings and generate the hyperlinks to the source code.

    Args:
        AbstractUtilities (class): Builder class of the abstract staticmethods of `Utilities`.
    """

    @staticmethod
    def insert_in_file(markdown_text: str, file_path: Path) -> None:
        """Insert the markdown formatted text into a new or existing file.

        Args:
            markdown_text (str): Text as string, which follows the markdown format.
            file_path (Path): Filename and path as Path object.

        Raises:
            RuntimeError: If {{autogenerated}} is not provided in a template file, it will cause a
                a **RuntimeError**.
        """
        if file_path.exists():
            template = file_path.read_text(encoding="utf-8")
            if "{{autogenerated}}" not in template:
                raise RuntimeError(
                    f"Template found for {file_path} but missing "
                    f"{{autogenerated}} tag."
                )
            markdown_text = template.replace("{{autogenerated}}", markdown_text)
            print(f"...inserting autogenerated content into template:{file_path}")
        else:
            print(f"...creating new page with autogenerated content:{file_path}")
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.write_text(markdown_text, encoding="utf-8")

    @staticmethod
    def element_to_mkdocstrings(element: str, titles_size: str) -> str:
        """Converts point separated string into the mkdocstrings format.

        For converting the elements to mkdocstrings, the element will added **:::** in front of the
         element string. In addition to that, the the new mkdocstrings will get subheadings.

        Args:
            element (str): String of they python class, function, or method, which has to be
                 converted to a string in the mkdocstrings format.
            titles_size (str): Current title size in the style of '#', which defines the headings.

        Returns:
            str: String of they python class, function, or method, which is converted to a string
                 in the mkdocstrings format.
        """
        return f"##{titles_size} :::{element}\n"

    @staticmethod
    def make_source_link(
        cls: classmethod,
        project_url: Union[str, Dict[str, str]],
        source: str = "**source code**",
    ) -> str:
        """Make a source link to the code basis including the linestart.

        Args:
            cls (classmethod): Convert a function to be a class method.
            project_url (Union[str, Dict[str, str]]): URL to the repository like GitHub
                 https://github.com/AI2Business/mkdocstrings-sourcelink/.
            source (str, optional): Name or sticker name for rendering the link to the
                 source.

        Returns:
            str: Hyperlink in html format with link to the repository.

        !!! tip "About *source*"
            Instead of using a string for `source = "**source code**"`, icons can be used instead
             or as combination of string + icon(s) like.
            ```python
                source = ":material-github::material-source-branch: source-code"
            ```
            In case of using material-icons, please check https://pictogrammers.github.io/@mdi/font/5.4.55/
            and replace `mdl` by `material`.
        """
        if isinstance(project_url, dict):
            if isinstance(cls, property):
                base_module = cls.fget.__module__.split(".")[0]
            else:
                base_module = cls.__module__.split(".")[0]
            project_url = project_url[base_module]

        if isinstance(cls, property):
            path = cls.fget.__module__.replace(".", "/")
            line = inspect.getsourcelines(cls.fget)[-1]
        else:
            path = cls.__module__.replace(".", "/")
            line = inspect.getsourcelines(cls)[-1]

        return (
            f'<span style="float:right;">'
            f"[{source}]({project_url}/{path}.py#L{line})"
            f"</span>"
        )

    @staticmethod
    def make_title(cls: classmethod, titles_size: str, underline_title: bool) -> str:
        """Make the title of the class, function, or method.

        Args:
            cls (classmethod): Convert a function to be a class method. In case of class properties
                 `fget` is used to read out the name of the module.
            titles_size (str): Current title size in the style of '#', which defines the headings.

        Returns:
            str: The name of the class, function, or method in a markdown conformed title.
        """
        title_underline = "\n---\n" if underline_title else "\n"
        if isinstance(cls, property):
            return f"#{titles_size} {cls.fget.__name__}{title_underline}"
        return f"#{titles_size} {cls.__name__}{title_underline}"

    @staticmethod
    def import_object(element: str) -> object:
        """Import an object like class, function, or method from a string.

        Args:
            element (str): String of class, function, or method, which should be converted to an
                 object.

        Returns:
            object: Class, function, or method object for the giving element.
        """
        last_object_got = None
        seen_names = []
        for name in element.split("."):
            seen_names.append(name)
            try:
                last_object_got = importlib.import_module(".".join(seen_names))
            except ModuleNotFoundError:
                last_object_got = getattr(last_object_got, name)

        return last_object_got

    @staticmethod
    def return_as_Path(path: str = None) -> Optional[Path]:
        """Converts strings to Path of pathlib.

        Args:
            path (str, optional): String of a filename.

        Returns:
            Optional[Path]: Path object of the initial filename.
        """
        if path:
            return Path(path)
        return None



from pathlib import Path

from setuptools import find_packages, setup

this_file = Path(__file__).resolve()
readme = this_file.parent / "README.md"


setup(
    name="mkdocstrings_sourcelink",
    version="0.3.2",
    packages=find_packages(),
    package_data={"": ["README.md"]},
    url="https://ai2business.github.io/mkdocstrings-sourcelink/",
    download_url="https://github.com/AI2Business/mkdocstrings-sourcelink",
    author="AI2Business",
    author_email="ai2business@protonmail.com",
    description="Automatic source link generation for mkdocstrings.",
    long_description=readme.read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    license="Apache License 2.0",
    extras_require={"tests": ["pytest", "pytest-cov"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
        "Topic :: Documentation",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
    ],
    keywords="markdown, mkdocs, mkdocstrings, docstrings, auto-doc, development",
    python_requires=">=3.6, <4",
    project_urls={
        "Bug Reports": "https://github.com/AI2Business/mkdocstrings-sourcelink/issues",
        "Source": "https://github.com/AI2Business/mkdocstrings-sourcelink/tree/main/mkdocstrings_sourcelink",
    },
)
